import os
import re
import shutil
import tempfile

from ..converter import KnowledgePostConverter


class DocxConverter(KnowledgePostConverter):

    _registry_keys = ['docx']

    @property
    def dependencies(self):
        # Dependencies required for this converter on top of core knowledge-repo dependencies
        return ['pypandoc']

    def from_file(self, filename, **opts):
        self.tmp_dir = wd = tempfile.mkdtemp()
        target_file = os.path.join(wd, 'post.md')

        import pypandoc
        pypandoc.convert_file(
            filename,
            format='docx',
            to='markdown-grid_tables',
            outputfile=target_file,
            extra_args=[
                '--standalone',
                '--wrap=none',
                '--extract-media={}'.format(wd)
            ]
        )

        with open(target_file) as f:
            md = f.read()

        # Image embeddings exported from docx files have fixed sizes in inches
        # which browsers do not understand. We remove these annotations.
        md = re.sub(r'(\!\[[^\]]+?\]\([^\)]+?\))\{[^\}]+?\}', lambda m: m.group(1), md)

        # Write markdown content to knowledge post (images will be extracted later)
        self.kp_write(md)

    def cleanup(self):
        if hasattr(self, 'tmp_dir'):
            shutil.rmtree(self.tmp_dir)
            del self.tmp_dir



""" gunicorn.py

    Utilities for running the knowledge app via gunicorn.

    Adapted from example in http://docs.gunicorn.org/en/stable/custom.html.
"""

from __future__ import absolute_import

from gunicorn.app.base import BaseApplication

from .common import KnowledgeDeployer


class GunicornDeployer(BaseApplication, KnowledgeDeployer):

    _registry_keys = ['gunicorn']

    def __init__(self, *args, **kwargs):
        KnowledgeDeployer.__init__(self, *args, **kwargs)
        BaseApplication.__init__(self)

    def load_config(self):
        env_args = self.cfg.parser().parse_args(self.cfg.get_cmd_args_from_env())

        # Load up environment configuration.
        for key, value in vars(env_args).items():
            if key != 'args' and value is not None:
                self.cfg.set(key, value)

        # Update the configuration with the options specified via KnowledgeDeployer
        options = {
            'bind': u'{}:{}'.format(self.host, self.port),
            'workers': self.workers,
            'timeout': self.timeout,
        }
        if self.app.config['DEPLOY_HTTPS']:
            options['certfile'] = self.app.config['SSL_CERT']['cert']
            options['keyfile'] = self.app.config['SSL_CERT']['key']
        for key, value in options.items():
            self.cfg.set(key, value)

    def load(self):
        if not self.app.check_thread_support():
            raise RuntimeError("Database configuration is not suitable for deployment (not thread-safe).")
        return self.app.start_indexing()



"""Add groups + permissions to posts

Revision ID: 36baffc5df12
Revises: 1b385158fd32
Create Date: 2016-11-02 15:37:12.808143

"""

# revision identifiers, used by Alembic.
revision = '36baffc5df12'
down_revision = '1b385158fd32'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table('groups',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(length=128), nullable=True),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('name')
    )
    op.create_table('assoc_group_user',
    sa.Column('group_id', sa.Integer(), nullable=True),
    sa.Column('user_id', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['group_id'], ['groups.id'], ),
    sa.ForeignKeyConstraint(['user_id'], ['users.id'], )
    )
    op.create_table('assoc_post_group',
    sa.Column('post_id', sa.Integer(), nullable=True),
    sa.Column('group_id', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['group_id'], ['groups.id'], ),
    sa.ForeignKeyConstraint(['post_id'], ['posts.id'], )
    )
    op.add_column(u'posts', sa.Column('private', sa.Integer(), nullable=True))


def downgrade():
    op.drop_column(u'posts', 'private')
    op.drop_table('assoc_post_group')
    op.drop_table('assoc_group_user')
    op.drop_table('groups')



from flask_principal import Permission

from . import roles

# Site roles and permissions
admin = Permission(roles.admin)
stats_view = Permission(roles.stats_view)

# Index permissions
index_view = Permission(roles.index_view)

# Post permissions
post_comment = Permission(roles.post_comment)
post_edit = Permission(roles.post_edit)
post_view = Permission(roles.post_view)
post_download = Permission(roles.post_download)



import sys
sys.path.insert(0, '..')

import threading
import time
from selenium.webdriver import Chrome
import knowledge_repo

def take_screenshots():
    driver = Chrome()
    driver.get('http://127.0.0.1:7000')
    driver.set_window_size(1260, 800)
    driver.save_screenshot('main.png')

    driver.find_element_by_css_selector('.feed-post').click()
    time.sleep(5)  # Wait for MathJax to do its thing.
    driver.save_screenshot('post.png')
    driver.close()

threading.Timer(1.25, lambda: take_screenshots()).start()

(
    knowledge_repo.KnowledgeRepository.for_uri('tests/test_repo')
    .get_app(POLICY_ANONYMOUS_DOWNLOADS=True)
    .run(port=7000)
)



import errno
import os
import shutil
import tempfile

import git

from knowledge_repo._version import __git_uri__


def clone_kr_to_directory(dir):
    dir = os.path.expanduser(dir)
    if not os.path.exists(dir):
        os.makedirs(dir)
    assert os.path.isdir(dir)

    try:
        repo = git.Repo(dir)
        repo.remote().fetch()
    except git.InvalidGitRepositoryError:
        repo = git.Repo.clone_from(__git_uri__, dir)


def checkout_revision_to_dir(repo_path, revision, dir):
    repo_path = os.path.expanduser(repo_path)
    dir = os.path.expanduser(dir)
    repo = git.Repo(repo_path)
    repo.remote().fetch()
    repo.git.checkout(revision)
    return repo.git.checkout_index('-a', '-f', prefix=dir)


class CheckedOutRevision(object):

    def __init__(self, repo_path, revision):
        self.repo_path = repo_path
        self.revision = revision

    def __enter__(self):
        self.dir = tempfile.mkdtemp()
        checkout_revision_to_dir(self.repo_path, self.revision, self.dir + '/')
        return self.dir

    def __exit__(self, exc_type, exc_value, traceback):
        shutil.rmtree(self.dir)



from flask import flash, redirect, render_template, request, current_app, Blueprint, url_for
from flask_login import login_user, logout_user, login_required
from flask_principal import identity_changed, AnonymousIdentity, PermissionDenied

from ..models import User
from ..proxies import current_app
from ..utils.auth import is_safe_url

blueprint = Blueprint('auth', __name__,
                      template_folder='../templates', static_folder='../static')


@blueprint.route('/auth/login', methods=['GET', 'POST'])
def login():

    providers = current_app.auth_providers

    if len(providers) == 1:
        return redirect(url_for('auth_provider.{}.prompt'.format(providers[0].name)))

    return render_template(
        'auth-switcher.html',
        providers=[{'name': provider.name, 'icon_uri': provider.icon_uri, 'link_text': provider.link_text} for provider in providers]
    )


@blueprint.route("/auth/logout")
@login_required
def logout():
    logout_user()

    # Notify flask principal that the user has logged out
    identity_changed.send(current_app._get_current_object(),
                          identity=AnonymousIdentity())

    return redirect(url_for('index.render_feed'))



import posixpath
import json
from flask import request, redirect
from six.moves.urllib.parse import urljoin
from ..models import User
from ..auth_provider import KnowledgeAuthProvider


PRESETS = {
    'oauth2': {},  # allows generic OAuth2 to be configured in server_config.py
    'bitbucket': {
        'base_url': 'https://api.bitbucket.org/2.0/',
        'authorization_url': 'https://bitbucket.org/site/oauth2/authorize',
        'token_url': 'https://bitbucket.org/site/oauth2/access_token',
        'auto_refresh_url': 'https://bitbucket.org/site/oauth2/access_token',
        'scopes': 'account',
        'user_info_endpoint': 'user',
        'user_info_mapping': {
            'identifier': 'username',
            'name': 'display_name',
            'avatar_uri': ['links', 'avatar', 'href']
        }
    },
    'github': {
        'base_url': 'https://api.github.com/',
        'authorization_url': 'https://github.com/login/oauth/authorize',
        'token_url': 'https://github.com/login/oauth/access_token',
        'auto_refresh_url': None,
        'scopes': None,
        'user_info_endpoint': 'user',
        'user_info_mapping': {
            'identifier': ['email', 'login'],
            'name': 'name',
            'avatar_uri': 'avatar_url'
        }
    },
    'google': {
        'base_url': 'https://www.googleapis.com/oauth2/v2/',
        'authorization_url': 'https://accounts.google.com/o/oauth2/auth',
        'token_url': 'https://accounts.google.com/o/oauth2/token',
        'auto_refresh_url': 'https://accounts.google.com/o/oauth2/token',
        'scopes': [
            'https://www.googleapis.com/auth/userinfo.email',
            'https://www.googleapis.com/auth/userinfo.profile'
        ],
        'user_info_endpoint': 'userinfo',
        'user_info_mapping': {
            'identifier': 'email',
            'name': 'name',
            'avatar_uri': 'picture'
        }
    }
}


def _resolve_oauth_config(name, local_namespace, config, *variables):
    presets = PRESETS.get(name, {})
    output = []
    for variable in variables:
        value = local_namespace.get(variable, None)
        if value is None:
            value = config.get('OAUTH_{}_{}'.format(name, variable).upper(), None)
        if value is None:
            value = presets.get(variable, None)
        output.append(value)
    return output


class OAuth2Provider(KnowledgeAuthProvider):
    _registry_keys = list(PRESETS.keys())

    def init(self,
             authorization_url=None,
             token_url=None,
             auto_refresh_url=None,
             scopes=None,
             user_info_endpoint=None,
             client_id=None,
             client_secret=None,
             user_info_mapping=None,
             verify_ssl_certs=None,
             validate=None):

        (self.base_url,
         self.authorization_url,
         self.token_url,
         self.auto_refresh_url,
         self.scopes,
         self.user_info_endpoint,
         self.client_id,
         self.client_secret,
         self.user_info_mapping,
         self.verify_ssl_certs,
         validate) = _resolve_oauth_config(
            self.name,
            locals(),
            self.app.config,
            'base_url',
            'authorization_url',
            'token_url',
            'auto_refresh_url',
            'scopes',
            'user_info_endpoint',
            'client_id',
            'client_secret',
            'user_info_mapping',
            'verify_ssl_certs',
            'validate'
        )
        if validate is not None:
            self.validate = lambda x: validate(self, x)

        if self.verify_ssl_certs is None:
            self.verify_ssl_certs = True

        host = self.app.config['SERVER_NAME'] or 'localhost:7000'
        scheme = self.app.config['PREFERRED_URL_SCHEME'] or 'http'
        redirect_url = '{}://{}'.format(scheme, host)
        redirect_path = '/auth/login/{}/authorize'.format(self.name)
        if self.app.config['APPLICATION_ROOT']:
            redirect_path = posixpath.join(self.app.config['APPLICATION_ROOT'], redirect_path)
        redirect_uri = urljoin(redirect_url, redirect_path)

        # Import OAuth deps here so we do not have a hard dependency on them
        from requests_oauthlib import OAuth2Session
        if self.client_id is None or self.client_secret is None:
            raise ValueError('You must configure a client id and client secret in order to use oauth')
        self.oauth_client = OAuth2Session(
            client_id=self.client_id,
            scope=self.scopes,
            redirect_uri=redirect_uri,
            auto_refresh_url=self.auto_refresh_url
        )

    def prompt(self):
        return redirect(self.oauth_client.authorization_url(self.authorization_url)[0])

    def get_endpoint_url(self, endpoint):
        return urljoin(self.base_url, endpoint)

    def get_user(self):
        self.oauth_client.state = request.args.get('state')
        self.oauth_client.fetch_token(
            self.token_url,
            client_secret=self.client_secret,
            code=request.args.get('code'),
            verify=self.verify_ssl_certs)
        return self.extract_user_from_api()

    def extract_user_from_api(self):

        def extract_from_dict(d, key):
            if isinstance(key, (list, tuple)):
                if len(key) == 1:
                    key = key[0]
                else:
                    return extract_from_dict(d[key[0]], key[1:])
            if isinstance(key, str):
                return d[key]
            raise RuntimeError("Invalid key type: {}.".format(key))

        response = self.oauth_client.get(self.get_endpoint_url(self.user_info_endpoint), verify=self.verify_ssl_certs)
        try:
            response_dict = json.loads(response.content)
            identifier = extract_from_dict(response_dict, self.user_info_mapping['identifier'])
            if identifier is None:
                raise ValueError("identifier '{}' not found in authentication response".format(self.user_info_mapping['identifier']))
            user = User(identifier=identifier)
            if 'name' in self.user_info_mapping:
                user.name = extract_from_dict(response_dict, self.user_info_mapping['name'])
            if 'avatar_uri' in self.user_info_mapping:
                user.avatar_uri = extract_from_dict(response_dict, self.user_info_mapping['avatar_uri'])
        except Exception:
            raise RuntimeError("Failure to extract user information from:\n\n {}".format(response.content))

        return user
